# coding: utf-8
"""
    Function set to compute LD correlation from a reference panel
    in predefined Region

    LD matrix are then transformed to the pandas sparse format
"""

import scipy as sc
import pandas as pd
import subprocess as sub
import pkg_resources
import numpy as np
import re


def launch_plink_ld(startpos, endpos, chr, reffile, folder):
    """
    launch plink linkage desiquilibrium correlation and save
    the ouput

    Args:
        startpos (int): position of the start of the window
        endpos (int): position of the end of the window
        chr (str): chromosome position
        reffile (str): reference panel file
        folder (str): output folder
    """
    bimref = reffile + ".bim"
    ref_panel = pd.read_csv(bimref, sep="\t", names=['chr', "nothing", 'pos', 'Ref_all', 'alt_all'], index_col = 1)

    ref_panel = ref_panel.loc[(ref_panel.pos > startpos) & (ref_panel.pos < endpos)]
    ref_panel.index.to_series().to_csv("./snp_list.txt", index=False)

    fo = "{0}/chr{1}_{2}_{3}".format(folder, chr, int(startpos), int(endpos))

    cmd = "plink --bfile {0} --r --ld-snp-list ./snp_list.txt --ld-window 50 --ld-window-kb 3000 --chr {1} --out {2}".format(reffile, chr, fo)
    print(cmd)

    try:
        sub.check_output(cmd, stderr=sub.STDOUT, shell=True)
    except sub.CalledProcessError as e:
        print(e.output)

    ld_results = pd.read_csv('{0}.ld'.format(fo), delim_whitespace=True)
    ld_results.loc[ld_results.R**2 > 0.4].to_csv('{0}.ld'.format(fo), index=False, sep="\t")


def generate_sparse_matrix(plink_ld, ref_chr_df, ld_file_type = "plink"):
    """
    Extract correlation matrix from the plink correlation
    file generated by ld_matrix.launch_plink_ld
    read plink results create a sparse dataframe LD-matrix
    then save it to a zipped pickle

    Can also load ld from scipy matrix

    Args:
        plink_ld (str): path to the plink correlation matrix file
        ref_chr_df (str):

    Returns:
        pandas.SparseDataFrame : Linkage desiquilibrium matrix
    """
    if ld_file_type=="plink":
        return load_plink_ld(plink_ld, ref_chr_df)
    else:
        if ld_file_type=="scipy":
            return load_sparse_matrix(plink_ld, ref_chr_df)
        else:
            raise ValueError("ld files must be plink or scipy sparse matrix")

def load_plink_ld(plink_ld, ref_chr_df):
    plink_ld = plink_ld + ".ld"
    plink_ld = pd.read_csv(plink_ld, sep = "\s+")
    mat_ld = plink_ld.pivot(index='SNP_A', columns='SNP_B', values='R')
    un_index = mat_ld.index.intersection(mat_ld.columns)
    mat_ld = mat_ld.reindex(index=un_index, columns=un_index)
    mat_ld.fillna(0, inplace=True)
    sym = np.where(np.abs(mat_ld.values) > np.abs(mat_ld.values.transpose()), mat_ld.values, mat_ld.values.transpose())

    mat_ld = pd.DataFrame(sym, index=mat_ld.index, columns=mat_ld.columns)
    int_index = ref_chr_df.index.intersection(mat_ld.index)

    re_index = ref_chr_df.loc[int_index].sort_values(by="pos").index
    mat_ld = mat_ld.loc[re_index, re_index]

    return mat_ld

def load_sparse_matrix(path_sparse_LD, ref_chr_df):
    """
    Load matrix from LD matrix from scipy sparse file
    """
    fi_ld = path_sparse_LD + ".npz"
    fi_index = path_sparse_LD + ".csv"

    Sp_M = sc.sparse.load_npz(fi_ld)
    snp_index = pd.read_csv(fi_index, header=None)
    Sp_M = Sp_M.todense()
    mat_ld = (np.where(np.abs(Sp_M) > np.abs(Sp_M.transpose()), Sp_M, Sp_M.transpose()))
    mat_ld = pd.DataFrame(mat_ld, index= snp_index.iloc[:,0], columns=snp_index.iloc[:,0])

    valid_id = ref_chr_df.index.intersection(mat_ld.index)
    return(mat_ld.loc[valid_id, valid_id])


def generate_genome_matrices(region_files, reffolder, folder_output, prefix, suffix = "", separa=","):
    """
    go through region files and compute LD matrix for each transform and
    save the results in a pandas sparse dataframe

    Args:
        region_files (str) : region file containing beginning and end position
        reffolder (str) : folder of reference panel
        folder_output (str): folder to save plink LD correlation result files
        separa (str): separateur used in region_files
        prefix (str): part of file name in reffolder before the chromosome number
        suffix (str): part of file name in reffolder after the chromosome number (without extension)
    """
    regions = pd.read_csv(region_files,sep=separa, dtype={'chr':str, 'start':np.int64, 'stop':np.int64})

    for reg in regions.iterrows():
        # input reference panel file
        fi_ref = "{0}/{1}{2}{3}".format(reffolder,prefix, reg[1]['chr'], suffix)
        #print(fi_ref)
        chr_int = re.search('([0-9X]{1,2})', str(reg[1]['chr'])).group()

        # Compute the LD correlation with LD
        launch_plink_ld(reg[1]['start'], reg[1]['stop'], chr_int, fi_ref, folder_output)

        #fi_plink = "{0}/{1}_{2}_{3}.ld".format(folder_output, reg[1]['chr'], reg[1]['startpos'], reg[1]["endpos"])
        #fo_mat = "{0}/{1}_{2}_{3}.mat".format(folder_output, reg[1]['chr'], reg[1]['startpos'], reg[1]["endpos"])
        #transform plink output to a compressed generate_sparse_matrix
        #generate_sparse_matrix(fi_plink, fo_mat)
